"""
Export neuroimaging results created by neuroimaging software packages
(FSL, AFNI, ...) following NIDM-Results specification.

Specification: http://nidm.nidash.org/specs/nidm-results.html

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2013-2014
"""

from prov.model import ProvBundle, ProvDocument
import rdflib
import os
import datetime
from nidmresults.objects.constants import *
from nidmresults.objects.modelfitting import *
from nidmresults.objects.contrast import *
from nidmresults.objects.inference import *
import uuid
import csv
import tempfile
import zipfile


class NIDMExporter():

    """
    Generic class to parse a result directory to extract the pieces of
    information to be stored in NIDM-Results and to generate a NIDM-Results
    export.
    """

    def __init__(self, version, out_dir, zipped=True):
        out_dirname = os.path.basename(out_dir)
        out_path = os.path.dirname(out_dir)

        # Create output path from output name
        self.zipped = zipped
        if not self.zipped:
            out_dirname = out_dirname+".nidm"
        else:
            out_dirname = out_dirname+".nidm.zip"
        out_dir = os.path.join(out_path, out_dirname)

        # Quit if output path already exists and user doesn't want to overwrite
        # it
        if os.path.exists(out_dir):
            msg = out_dir+" already exists, overwrite?"
            if not raw_input("%s (y/N) " % msg).lower() == 'y':
                quit("Bye.")
            if os.path.isdir(out_dir):
                shutil.rmtree(out_dir)
            else:
                os.remove(out_dir)
        self.out_dir = out_dir

        # A temp directory that will contain the exported data
        self.export_dir = tempfile.mkdtemp(prefix="nidm-", dir=out_path)

        if version == "dev":
            self.version = {'major': 10000, 'minor': 0, 'revision': 0,
                            'num': version}
        else:
            major, minor, revision = version.split(".")
            if "-rc" in revision:
                revision, rc = revision.split("-rc")
            else:
                rc = -1
            self.version = {'major': int(major), 'minor': int(minor),
                            'revision': int(revision), 'rc': int(rc),
                            'num': version}

        # Initialise prov document
        self.doc = ProvDocument()
        self._add_namespaces()

        # Initialise rdf document
        self.g = rdflib.Graph()
        self.g.bind("nidm", "http://purl.org/nidash/nidm#")
        self.g.bind("niiri", "http://iri.nidash.org/")
        self.g.bind(
            "crypto",
            "http://id.loc.gov/vocabulary/preservation/cryptographicHashFuncti\
ons#")
        self.g.bind("fsl", "http://purl.org/nidash/fsl#")
        self.g.bind("spm", "http://purl.org/nidash/spm#")
        self.g.bind("dct", "http://purl.org/dc/terms/")
        self.g.bind("obo", "http://purl.obolibrary.org/obo/")
        self.g.bind("dctype", "http://purl.org/dc/dcmitype/")
        self.g.bind("nlx", "http://neurolex.org/wiki/")
        self.g.bind("prov", "http://www.w3.org/ns/prov#")
        self.g.bind("dc", "http://purl.org/dc/elements/1.1/")
        self.g.bind(
            "nfo",
            "http://www.semanticdesktop.org/ontologies/2007/03/22/nfo#")

    def parse(self):
        """
        Parse a result directory to extract the pieces information to be
        stored in NIDM-Results.
        """
        # Methods: find_software, find_model_fitting, find_contrasts and
        # find_inferences should be defined in the children classes and return
        # a list of NIDM Objects as specified in the objects module

        # Object of type Software describing the neuroimaging software package
        # used for the analysis
        self.software = self._find_software()

        # List of objects (or dictionary) of type ModelFitting describing the
        # model fitting step in NIDM-Results (main activity: Model Parameters
        # Estimation)
        self.model_fittings = self._find_model_fitting()

        # Dictionary of (key, value) pairs where where key is a tuple
        # containing the identifier of a ModelParametersEstimation object and a
        # tuple of identifiers of ParameterEstimateMap objects and value is an
        # object of type Contrast describing the contrast estimation step in
        # NIDM-Results (main activity: Contrast Estimation)
        self.contrasts = self._find_contrasts()

        # Inference activity and entities
        # Dictionary of (key, value) pairs where key is the identifier of a
        # ContrastEstimation object and value is an object of type Inference
        # describing the inference step in NIDM-Results (main activity:
        # Inference)
        self.inferences = self._find_inferences()

    def add_object(self, nidm_object):
        """
        Add a NIDMObject to a NIDM-Results export.
        """
        nidm_object.export(self.version)
        # ProvDocument: add object to the bundle
        self.bundle.update(nidm_object.p)
        # RDF document: add object to the main graph
        self.g += nidm_object.g

    def export(self):
        """
        Generate a NIDM-Results export.
        """
        if not os.path.isdir(self.export_dir):
            os.mkdir(self.export_dir)

        # Initialise main bundle
        self._create_bundle(self.version)
        self.add_object(self.software)

        # Add model fitting steps
        for model_fitting in self.model_fittings.values():
            model_fitting.activity.wasAssociatedWith(self.software)
            self.add_object(model_fitting)

        # Add contrast estimation steps
        for (model_fitting_id, pe_ids), contrasts in self.contrasts.items():
            model_fitting = self._get_model_fitting(model_fitting_id)
            for contrast in contrasts:
                contrast.estimation.used(model_fitting.rms_map)
                contrast.estimation.used(model_fitting.mask_map)
                contrast.estimation.wasAssociatedWith(self.software)

                for pe_id in pe_ids:
                    contrast.estimation.used(pe_id)

                self.add_object(contrast)

        # Add inference steps
        for contrast_id, inferences in self.inferences.items():
            contrast = self._get_contrast(contrast_id)

            for inference in inferences:
                if contrast.z_stat_map:
                    used_id = contrast.z_stat_map.id
                else:
                    used_id = contrast.stat_map.id
                inference.inference_act.used(used_id)
                inference.inference_act.wasAssociatedWith(self.software)

                self.add_object(inference)

        # Write-out prov file
        self.save_prov_to_files()

        return self.out_dir

    def _get_model_fitting(self, mf_id):
        """
        Retreive model fitting with identifier 'mf_id' from the list of model
        fitting objects stored in self.model_fitting
        """
        for model_fitting in self.model_fittings.values():
            if model_fitting.activity.id == mf_id:
                return model_fitting
        raise Exception("Model fitting activity with id: " + str(mf_id) +
                        " not found.")

    def _get_contrast(self, con_id):
        """
        Retreive contrast with identifier 'con_id' from the list of contrast
        objects stored in self.contrasts
        """
        for contrasts in self.contrasts.values():
            for contrast in contrasts:
                if contrast.estimation.id == con_id:
                    return contrast
        raise Exception("Contrast activity with id: " + str(con_id) +
                        " not found.")

    def _add_namespaces(self):
        """
        Add namespaces to NIDM document.
        """
        self.doc.add_namespace(NIDM)
        self.doc.add_namespace(NIIRI)
        self.doc.add_namespace(CRYPTO)
        self.doc.add_namespace(DCT)

    def _create_bundle(self, version):
        """
        Initialise NIDM-Results bundle.
        """
        # *** Bundle entity
        bundle_id = NIIRI[str(uuid.uuid4())]

        # provn export
        self.bundle = ProvBundle(identifier=bundle_id)

        self.doc.entity(bundle_id,
                        other_attributes=((PROV['type'], PROV['Bundle'],),
                                          (PROV['label'], "NIDM-Results"),
                                          (PROV['type'], NIDM_RESULTS),
                                          (NIDM_VERSION, version['num']))
                        )

        # ttl export
        bundle_uri = rdflib.URIRef(bundle_id.uri)
        self.g.add((
            bundle_uri, RDF.type, rdflib.URIRef(PROV["Bundle"].uri)))
        self.g.add((
            bundle_uri, RDF.type, rdflib.URIRef(PROV["Entity"].uri)))
        self.g.add((
            bundle_uri, RDF.type, rdflib.URIRef(NIDM_RESULTS.uri)))
        self.g.add((
            bundle_uri, RDFS.label, rdflib.Literal("NIDM-Results")))
        self.g.add((
            bundle_uri,
            rdflib.URIRef(NIDM_VERSION.uri),
            rdflib.Literal(version['num'], datatype=XSD.string)))

        # *** NIDM-Results Export Activity
        if version['num'] not in ["1.0.0", "1.1.0"]:
            self.export = NIDMResultsExport()
            self.export.export(self.version)
            self.doc.update(self.export.p)
            self.g += self.export.g

        # *** bundle was Generated by NIDM-Results Export Activity
        ctime = datetime.datetime.now().time()
        # Qualified generation
        bnode = rdflib.BNode()
        self.g.add((
            bnode, RDF.type, rdflib.URIRef(PROV["Generation"].uri)))
        self.g.add((
            bnode,
            rdflib.URIRef(PROV["atTime"].uri),
            rdflib.Literal(str(ctime), datatype=XSD.dateTime)))
        self.g.add((
            bundle_uri,
            rdflib.URIRef(PROV["qualifiedGeneration"].uri), bnode))

        if version['num'] in ["1.0.0", "1.1.0"]:
            self.doc.wasGeneratedBy(bundle_id, time=str(ctime))
        else:
            # provn
            self.bundle.wasGeneratedBy(self.export.id, time=str(ctime))
            # ttl
            self.g.add((
                bnode,
                rdflib.URIRef(PROV['activity'].uri),
                rdflib.URIRef(self.export.id.uri)))

        # *** NIDM-Results Exporter (Software Agent)
        if version['num'] not in ["1.0.0", "1.1.0"]:
            self.exporter = self._get_exporter()
            self.exporter.export(self.version)
            self.doc.update(self.exporter.p)
            self.g += self.exporter.g

            self.exporter.wasAssociatedWith(self.export.id)
            self.g.add((
                rdflib.URIRef(self.export.id.uri),
                rdflib.URIRef(PROV['wasAssociatedWith'].uri),
                rdflib.URIRef(self.exporter.id.uri)))

    def _get_model_parameters_estimations(self, error_model):
        """
        Infer model estimation method from the 'error_model'. Return an object
        of type ModelParametersEstimation.
        """
        if error_model.dependance == NIDM_INDEPEDENT_ERROR:
            if error_model.variance_homo:
                estimation_method = STATO_OLS
            else:
                estimation_method = STATO_WLS
        else:
            estimation_method = STATO_GLS

        mpe = ModelParametersEstimation(estimation_method, self.software.id)

        return mpe

    def use_prefixes(self, ttl):
        prefix_file = os.path.join(os.path.dirname(__file__), 'prefixes.csv')
        with open(prefix_file, 'rb') as csvfile:
            reader = csv.reader(csvfile)
            for alphanum_id, prefix, uri in reader:
                if alphanum_id in ttl:
                    ttl = "@prefix " + prefix + ": <" + uri + "> .\n" + ttl
                    ttl = ttl.replace(alphanum_id, prefix + ":")
        return ttl

    def save_prov_to_files(self, showattributes=False):
        """
        Write-out provn serialisation to nidm.provn.
        """
        self.doc.add_bundle(self.bundle)
        # provn_file = os.path.join(self.export_dir, 'nidm.provn')
        # provn_fid = open(provn_file, 'w')
        # # FIXME None
        # provn_fid.write(self.doc.get_provn(4).replace("None", "-"))
        # provn_fid.close()

        ttl_file = os.path.join(self.export_dir, 'nidm.ttl')
        ttl_txt = self.g.serialize(format='turtle')

        ttl_txt = self.use_prefixes(ttl_txt)
        with open(ttl_file, 'w') as ttl_fid:
            ttl_fid.write(ttl_txt)

        # Post-processing
        if not self.zipped:
            # Just rename temp directory to output_path
            os.rename(self.export_dir, self.out_dir)
        else:
            # Create a zip file that contains the content of the temp directory
            os.chdir(self.export_dir)
            zf = zipfile.ZipFile(os.path.join("..", self.out_dir), mode='w')
            try:
                for root, dirnames, filenames in os.walk("."):
                    for filename in filenames:
                        zf.write(os.path.join(filename))
                shutil.rmtree(os.path.join("..", self.export_dir))
            finally:
                zf.close()
                os.chdir("..")

        # ttl_fid = open(ttl_file, 'w');
        # serialization is done in xlm rdf
        # graph = Graph()
        # graph.parse(data=self.doc.serialize(format='rdf'), format="xml")
        # ttl_fid.write(graph.serialize(format="turtle"))
        # ttl_fid.write(self.doc.serialize(format='rdf').
            # replace("inf", '"INF"'))
        # ttl_fid.close()
        # print "provconvert -infile " + provn_file + " -outfile " + ttl_file
        # check_call("provconvert -infile " + provn_file +
        #            " -outfile " + ttl_file, shell=True)



"""
Generic objects supporting the classes defined in NIDM-Results.

Specification: http://nidm.nidash.org/specs/nidm-results.html

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2013-2014
"""
from prov.model import ProvBundle, Identifier
import prov
import numpy as np
import os
from constants import *
import nibabel as nib
import shutil
import hashlib
import uuid
import rdflib
import json
from rdflib.namespace import RDF, RDFS, XSD


class NIDMObject(object):

    """
    Generic class, parent of all objects describing a NIDM entity, activity
    or agent
    """

    def __init__(self, export_dir=None, oid=None):
        self.export_dir = export_dir
        self.p = ProvBundle()

        self.g = rdflib.Graph()
        if oid is None:
            self.id = NIIRI[str(uuid.uuid4())]
        else:
            self.id = oid

    def __str__(self):
        value = ""
        if hasattr(self, 'value'):
            value = ": " + self.value
        location = ""
        if hasattr(self, 'file'):
            if hasattr(self.file, 'path'):
                location = " - " + self.file.path
        return '"' + self.label + '"' + value + location

    def __repr__(self):
        return '<"' + self.label + '" ' + \
               str(self.id).replace("niiri:", "").replace(NIIRI._uri, "")[0:8]\
               + '>'

    def _rdf_add_attributes(self, attributes):
        if self.type is not None:
            self._rdf_add(self.id, RDF.type, self.type)
        self._rdf_add(self.id, RDF.type, self.prov_type)

        # If attributes is a dictionnary, convert to list
        if isinstance(attributes, dict):
            attributes = [(k, v) for k, v in attributes.iteritems()]

        for p, o in attributes:
            self._rdf_add(self.id, p, o)

    def _rdf_add(self, s, p, o):
        if isinstance(o, prov.identifier.QualifiedName):
            o = rdflib.URIRef(o.uri)
        elif isinstance(o, prov.identifier.Identifier):
            o = rdflib.Literal(o, datatype=XSD.anyURI)
        else:
            if (isinstance(o, str) or isinstance(o, unicode)) and \
                    not str(p) == "prov:label":
                o = rdflib.Literal(o, datatype=XSD.string)
            elif isinstance(o, float):
                o = rdflib.Literal(o, datatype=XSD.float)
            elif isinstance(o, bool):
                o = rdflib.Literal(o, datatype=XSD.boolean)
            elif isinstance(o, int):
                o = rdflib.Literal(o, datatype=XSD.int)
            else:
                o = rdflib.Literal(str(o))

        if not isinstance(p, rdflib.URIRef):
            if str(p) == "prov:type":
                p = RDF.type
            elif str(p) == "prov:label":
                p = RDFS.label
            elif str(p) == "prov:location":
                p = rdflib.URIRef(p.uri.replace("location", "atLocation"))
            else:
                p = rdflib.URIRef(p.uri)

        self.g.add((rdflib.URIRef(s.uri), p, o))

    def add_object(self, nidm_object, nidm_version):
        nidm_object.export(nidm_version)

        # Prov graph (=> provn)
        self.p.update(nidm_object.p)
        # RDF graph (=> turtle)
        self.g = self.g + nidm_object.g

    def used(self, nidm_object):
        self._add_prov_relation(PROV['used'], nidm_object)

    def wasGeneratedBy(self, nidm_object):
        self._add_prov_relation(PROV['wasGeneratedBy'], nidm_object)

    def wasDerivedFrom(self, nidm_object):
        self._add_prov_relation(PROV['wasDerivedFrom'], nidm_object)

    def wasAttributedTo(self, nidm_object):
        self._add_prov_relation(PROV['wasAttributedTo'], nidm_object)

    def wasAssociatedWith(self, nidm_object):
        self._add_prov_relation(PROV['wasAssociatedWith'], nidm_object)

    def _add_prov_relation(self, relation, nidm_object):
        if isinstance(nidm_object, NIDMObject):
            object_id = nidm_object.id
        else:
            object_id = nidm_object

        if relation == PROV['used']:
            self.p.used(self.id, object_id)
        elif relation == PROV['wasGeneratedBy']:
            self.p.wasGeneratedBy(self.id, object_id)
        elif relation == PROV['wasDerivedFrom']:
            self.p.wasDerivedFrom(self.id, object_id)
        elif relation == PROV['wasAssociatedWith']:
            self.p.wasAssociatedWith(self.id, object_id)
        elif relation == PROV['wasAttributedTo']:
            self.p.wasAttributedTo(self.id, object_id)
        else:
            raise Exception('Unrecognised prov relation')

        self._rdf_add(self.id, relation, object_id)

    def add_attributes(self, attributes):
        if self.prov_type == PROV['Activity']:
            self.p.activity(self.id, other_attributes=attributes)
        elif self.prov_type == PROV['Entity']:
            self.p.entity(self.id, other_attributes=attributes)
        elif self.prov_type == PROV['Agent']:
            self.p.agent(self.id, other_attributes=attributes)

        self._rdf_add_attributes(attributes)


# class NIDMBundle(NIDMObject):
#     """
#     Object representing a NIDM Bundle entity.
#     """

#     def __init__(self, version):
#         self.id = NIIRI[str(uuid.uuid4())]
#         self.version = version
#         self.type = NIDM_RESULTS
#         self.prov_type = PROV['Bundle']

#     def export(self):
#         self.bundle = ProvBundle(identifier=bundle_id)

#         self.doc.entity(bundle_id,
#                         other_attributes=((PROV['type'], PROV['Bundle'],),
#                                           (PROV['label'], "NIDM-Results"),
#                                           (PROV['type'], NIDM_RESULTS),
#                                           (NIDM_VERSION, version))
#                         )

#         self.doc.wasGeneratedBy(bundle_id,
#                                 time=str(datetime.datetime.now().time()))

class CoordinateSpace(NIDMObject):
    """
    Object representing a CoordinateSpace entity.
    """

    def __init__(self, coordinate_system, nifti_file=None, vox_to_world=None,
                 vox_size=None, dimensions=None, numdim=None, units=None,
                 oid=None, label="Coordinate space"):
        super(CoordinateSpace, self).__init__(oid)
        self.coordinate_system = coordinate_system
        self.type = NIDM_COORDINATE_SPACE
        self.prov_type = PROV['Entity']
        self.label = label

        if (vox_to_world is None) and (vox_size is None) and\
                (dimensions is None) and (numdim is None) and\
                (units is None) and \
                (nifti_file is not None):
            thresImg = nib.load(nifti_file)
            thresImgHdr = thresImg.get_header()

            numdim = len(thresImg.shape)
            dimensions = thresImg.shape
            # FIXME: is vox_to_world the qform?
            vox_to_world = thresImg.get_qform()
            vox_size = thresImgHdr['pixdim'][1:(numdim + 1)]
            # FIXME: this gives mm, sec => what is wrong: FSL file, nibabel,
            # other?
            # units = str(thresImgHdr.get_xyzt_units()).strip('()')
            units = ["mm", "mm", "mm"]

        self.number_of_dimensions = numdim
        self.voxel_to_world = vox_to_world
        self.voxel_size = vox_size
        self.dimensions = dimensions
        self.units = units

    def is_mni(self):
        if str(self.coordinate_system) in [
                NIDM_MNI_COORDINATE_SYSTEM.uri,
                NIDM_ICBM_MNI152_LINEAR_COORDINATE_SYSTEM.uri,
                NIDM_ICBM_MNI152_NON_LINEAR2009A_ASYMMETRIC_COORDINATE_SYSTEM.uri,
                NIDM_ICBM_MNI152_NON_LINEAR2009A_SYMMETRIC_COORDINATE_SYSTEM.uri,
                NIDM_ICBM_MNI152_NON_LINEAR2009B_ASYMMETRIC_COORDINATE_SYSTEM.uri,
                NIDM_ICBM_MNI152_NON_LINEAR2009B_SYMMETRIC_COORDINATE_SYSTEM.uri,
                NIDM_ICBM_MNI152_NON_LINEAR2009C_ASYMMETRIC_COORDINATE_SYSTEM.uri,
                NIDM_ICBM_MNI152_NON_LINEAR2009C_SYMMETRIC_COORDINATE_SYSTEM.uri,
                NIDM_ICBM_MNI152_NON_LINEAR6TH_GENERATION_COORDINATE_SYSTEM.uri,
                NIDM_ICBM452_AIR_COORDINATE_SYSTEM.uri,
                NIDM_ICBM452_WARP5_COORDINATE_SYSTEM.uri,
                NIDM_IXI549_COORDINATE_SYSTEM.uri,
                NIDM_MNI305_COORDINATE_SYSTEM.uri]:
            return True
        else:
            return False

    def is_talairach(self):
        if str(self.coordinate_system) in [NIDM_TALAIRACH_COORDINATE_SYSTEM.uri]:
            return True
        else:
            return False

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        self.add_attributes({
            PROV['type']: self.type,
            NIDM_DIMENSIONS_IN_VOXELS: json.dumps(self.dimensions),
            NIDM_NUMBER_OF_DIMENSIONS: self.number_of_dimensions,
            NIDM_VOXEL_TO_WORLD_MAPPING:
            json.dumps(self.voxel_to_world.tolist()),
            NIDM_IN_WORLD_COORDINATE_SYSTEM: self.coordinate_system,
            NIDM_VOXEL_UNITS: json.dumps(self.units),
            NIDM_VOXEL_SIZE: json.dumps(self.voxel_size.tolist()),
            PROV['label']: self.label})
        return self.p


class NIDMFile(NIDMObject):
    """
    Object representing a File (to be used as attribute of another class)
    """
    def __init__(self, rdf_id, location, new_filename=None, export_dir=None,
                 sha=None, format=None, temporary=False):
        super(NIDMFile, self).__init__(export_dir)
        self.prov_type = PROV['Entity']
        self.path = location
        if new_filename is None:
            # Keep same file name
            path, self.new_filename = os.path.split(self.path)
        else:
            self.new_filename = new_filename

        # NIDMFile is not a NIDM class defined in the owl file
        self.type = None
        self.id = rdf_id
        self.label = "'NIDM file'"  # used if display is called

        self.sha = sha
        self.format = format
        self.temporary = temporary

    def is_nifti(self):
        return self.path.endswith(".nii") or \
            self.path.endswith(".nii.gz") or \
            self.path.endswith(".img") or \
            self.path.endswith(".hrd")

    def get_sha_sum(self, nifti_file):
        nifti_img = nib.load(nifti_file)
        data = nifti_img.get_data()
        # Fix needed as in https://github.com/pymc-devs/pymc/issues/327
        if not data.flags["C_CONTIGUOUS"]:
            data = np.ascontiguousarray(data)

        return hashlib.sha512(data).hexdigest()

    def export(self, nidm_version):
        """
        Copy file over of export_dir and create corresponding triples
        """
        if self.path is not None:
            path, org_filename = os.path.split(self.path)
            if self.export_dir is not None:
                # Copy file only if export_dir is not None
                new_file = os.path.join(self.export_dir, self.new_filename)
                if not self.path == new_file:
                    shutil.copy(self.path, new_file)
                    if self.temporary:
                        os.remove(self.path)
            else:
                new_file = self.path

            if nidm_version['num'] in ["1.0.0", "1.1.0"]:
                loc = Identifier("file://./" + self.new_filename)
            else:
                loc = Identifier(self.new_filename)

            self.add_attributes([(NFO['fileName'], self.new_filename)])

            if self.export_dir:
                self.add_attributes([(PROV['atLocation'], loc)])

            if nidm_version['num'] in ("1.0.0", "1.1.0"):
                if org_filename is not self.new_filename:
                    self.add_attributes([(NFO['fileName'], org_filename)])

            if self.is_nifti():
                if self.sha is None:
                    self.sha = self.get_sha_sum(new_file)
                if self.format is None:
                    self.format = "image/nifti"

                self.add_attributes([
                    (CRYPTO['sha512'], self.sha),
                    (DCT['format'], self.format)
                ])

            return self.p


class Image(NIDMObject):

    """
    Object representing an Image entity.
    """

    def __init__(self, export_dir, image_file, filename):
        super(Image, self).__init__(export_dir)
        self.type = DCTYPE['Image']
        self.prov_type = PROV['Entity']
        self.id = NIIRI[str(uuid.uuid4())]
        self.file = NIDMFile(self.id, image_file, filename, export_dir)

    def export(self, nidm_version):
        """
        Create prov entity.
        """
        if self.file is not None:
            self.add_object(self.file, nidm_version)

            self.add_attributes([
                (PROV['type'], self.type),
                (DCT['format'], "image/png"),
            ])

        return self.p


class NeuroimagingSoftware(NIDMObject):
    """
    Class representing a NeuroimagingSoftware Agent.
    """

    def __init__(self, software_type, version):
        super(NeuroimagingSoftware, self).__init__()
        self.id = NIIRI[str(uuid.uuid4())]
        self.version = version
        # FIXME: get label from owl!
        if software_type.lower() == "fsl":
            self.name = "FSL"
        else:
            raise Exception('Unrecognised software: ' + str(software_type))
        self.type = SCR_FSL  # NLX_FSL
        self.prov_type = PROV['Agent']

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        if nidm_version['major'] < 1 or \
                (nidm_version['major'] == 1 and nidm_version['minor'] < 3):
            self.type = NLX_OLD_FSL

        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['type'], PROV['SoftwareAgent']),
            (PROV['label'], self.name),
            (NIDM_SOFTWARE_VERSION, self.version))
        )

        return self.p


class ExporterSoftware(NIDMObject):
    """
    Class representing a Software Agent.
    """

    def __init__(self, software_type, version):
        super(ExporterSoftware, self).__init__()
        self.id = NIIRI[str(uuid.uuid4())]
        self.type = software_type
        self.prov_type = PROV['Agent']
        self.version = version

        if software_type == NIDM_FSL:
            self.name = "nidmfsl"
        else:
            raise Exception('Unrecognised software: ' + str(software_type))

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['type'], PROV['SoftwareAgent']),
            (PROV['label'], self.name),
            (NIDM_SOFTWARE_VERSION, self.version))
        )

        return self.p


class NIDMResultsExport(NIDMObject):
    """
    Class representing a NIDM-Results Export activity.
    """
    def __init__(self):
        super(NIDMResultsExport, self).__init__()
        self.id = NIIRI[str(uuid.uuid4())]
        self.type = NIDM_NIDM_RESULTS_EXPORT
        self.prov_type = PROV['Activity']
        self.label = "NIDM-Results export"

    def export(self, nidm_version):
        """
        Create prov entities and activities.
        """
        self.add_attributes([(PROV['label'], self.label)])

        return self.p



#!/usr/bin/env python
'''Read an owl file and extract interesting information

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2014
'''

from rdflib import RDF, term
from rdflib.graph import Graph
from rdflib.term import Literal
from nidmresults.objects.constants_rdflib import *
import urllib2
import warnings
import vcr
import os
import logging
import csv
import collections
import re
import warnings

RELPATH = os.path.dirname(os.path.abspath(__file__))
NIDM_PATH = os.path.dirname(RELPATH)

logging.basicConfig(filename='debug.log', level=logging.DEBUG, filemode='w')
logger = logging.getLogger(__name__)


class OwlReader():

    def __init__(self, owl_file, import_owl_files=None):
        self.file = owl_file
        self.import_files = import_owl_files
        self.graph = self.get_graph()

        # Retreive all classes defined in the owl file
        self.classes = self.get_class_names()
        self.properties = self.get_property_names()
        self.individuals = self.get_individuals()
        # For each class find out attribute list as defined by domain in
        # attributes
        self.attributes, self.ranges, \
            self.type_restrictions, self.parent_ranges = self.get_attributes()

        labels = self.graph.subject_objects(RDFS['label'])
        # Ignore deprecated
        self.labels = collections.OrderedDict(
            (str(key), value) for (value, key) in labels
            if not self.is_deprecated(value))

    def get_class_names(self):
        # Add PROV sub-types
        sub_types = set([PROV['Bundle'], PROV['Location'],
                         PROV['Collection']])

        for class_name in self.graph.subjects(RDF['type'], OWL['Class']):
            if not isinstance(class_name, term.BNode):
                sub_types.add(class_name)

        return sub_types

    def get_direct_children(self, term):
        # Find all direct children of 'term'
        children = set()

        for class_name in self.graph.subjects(RDFS['subClassOf'], term):
            if not self.is_deprecated(class_name):
                children.add(class_name)

        return children

    def get_direct_parents(self, term):
        # Find all direct parents of 'term'
        parents = set()

        for parent_name in self.graph.objects(term, RDFS['subClassOf']):
            parents.add(parent_name)

        return parents

    def get_nidm_parent(self, term):
        # Find direct nidm parent of 'term'
        parents = self.get_direct_parents(term)

        for parent in parents:
            if not self.is_external_namespace(parent):
                return parent

        return None

    def is_class(self, uri):
        return (uri, RDF['type'], OWL['Class']) in self.graph

    def all_of_rdf_type(self, rdf_type, prefix=None, but=set(),
                        but_type=OWL['AnnotationProperty']):
        classes = self.graph.subjects(RDF['type'], rdf_type)

        deprecated = self.graph.subjects(
            OWL['deprecated'], term.Literal(True))
        but = set(but).union(set(deprecated))
        annotations = self.graph.subjects(RDF['type'], but_type)
        but = but.union(set(annotations))

        # FIXME: Is there a more efficient way?
        if prefix:
            original_classes = classes
            classes = list()
            for class_name in original_classes:
                if class_name.startswith(prefix):
                    classes.append(class_name)
        if but:
            classes = list(set(classes) - set(but))

        classes = sorted(classes)
        return classes

    def get_classes(self, prefix=None, but=None):
        return self.all_of_rdf_type(OWL['Class'], prefix, but)

    def get_by_namespaces(self, term_list, but=None):
        by_nsp = dict()
        ignored = []

        for uri in term_list:
            if not isinstance(uri, term.BNode):
                try:
                    qname = self.graph.qname(uri)
                    nsp = qname.split(":")[0]
                except:
                    qname = None
                    nsp = None

                if (str(nsp) not in but) and \
                        ((qname is None) or (not qname.startswith(but))):
                    by_nsp.setdefault(nsp, list()).append(uri)
                else:
                    ignored.append(uri)

        # if ignored:
        #     warnings.warn("Ignoring... " + "\n\t".join(ignored))

        return by_nsp

    def count_by_namespaces(self):
        owl_types = list([OWL['Class'], OWL['DatatypeProperty'],
                          OWL['ObjectProperty'], OWL['NamedIndividual']])

        # Ignore the following namespaces/terms (not part of the model) from
        # the count
        but = ("owl", "rdf", "prv", "protege", "xsd", "obo:IAO_", "iao",
               "obo:iao.owl", "prov")

        counter_dict = dict()

        counter = 0

        all_terms = dict()
        for owl_type in owl_types:
            terms = self.get_by_namespaces(self.all_of_rdf_type(owl_type), but)
            keys = set(all_terms).union(terms)
            no = []
            all_terms = dict((k, all_terms.get(k, no) + terms.get(k, no)) for k in keys)

            len_dict = {key: (len(value), terms) for (key, value) in \
                terms.items() if key is not None}
            num = sum([x[0] for x in len_dict.values()])
            type_id = self.graph.qname(owl_type).split(":")[1]
            counter = counter + num
            # print counter
            comp_to = ""

            counter_dict[type_id] = (num, len_dict)

        num_attributes = counter_dict[u'DatatypeProperty'][0] + \
            counter_dict[u'ObjectProperty'][0]
        num_classes = counter_dict[u'Class'][0] + \
            counter_dict[u'NamedIndividual'][0]
        num_terms = num_attributes + num_classes

        all_terms_len = {key: (len(value), value) for (key, value) in
                         all_terms.items() if key is not None}

        num_defined = all_terms_len[u'nidm'][0] + all_terms_len[u'spm'][0] + \
            all_terms_len[u'fsl'][0]
        num_reused = num_terms - num_defined

        # Sanity check
        num_terms_from_all = sum([x[0] for x in all_terms_len.values()])
        if not num_terms_from_all == num_terms:
            raise Exception('Error: number of terms from all is inconsistent.')

        return (num_terms, num_classes, num_attributes, num_reused,
                all_terms_len)

    def get_class_names_by_prov_type(self, classes=None, prefix=None,
                                     but=None):
        class_names = dict()
        # We at least want to have an output for Entity, Activity and Agent
        class_names[PROV['Entity']] = list()
        class_names[PROV['Activity']] = list()
        class_names[PROV['Agent']] = list()

        class_names[None] = list()

        if not classes:
            classes = self.get_classes(prefix, but)

        for class_name in classes:
            if not self.is_class(class_name):
                raise Exception('Class '+str(class_name)+' does not exist.')

            if not isinstance(class_name, term.BNode):
                prov_type = self.get_prov_class(class_name)
                if prov_type:
                    class_names.setdefault(prov_type, list())\
                               .append(class_name)
                else:
                    prov_type_found = False
                    parent_classes = self.get_direct_parents(class_name)
                    for parent_class in parent_classes:
                        prov_type = self.get_prov_class(parent_class,
                                                        recursive=3)
                        if prov_type:
                            class_names.setdefault(prov_type, list())\
                                       .append(class_name)
                            prov_type_found = True

                    if not prov_type_found:
                        warnings.warn('No PROV type for class: ' +
                                      self.graph.qname(class_name))
                        class_names.setdefault(None, list()).append(class_name)

        return class_names

    def is_deprecated(self, term):
        deprecated = False
        if (term, OWL['deprecated'], Literal(True)) in self.graph:
            deprecated = True
        return deprecated

    def get_property_names(self):
        properties = set()
        for class_name in self.graph.subjects(
                RDF['type'], OWL['DatatypeProperty']):
            if not self.is_deprecated(class_name):
                properties.add(class_name)
        for class_name in self.graph.subjects(
                RDF['type'], OWL['ObjectProperty']):
            if not self.is_deprecated(class_name):
                properties.add(class_name)
        return properties

    def get_individuals(self, uri=None):
        individuals = set()

        if uri is None:
            individuals = self.graph.subjects(
                RDF['type'], OWL['NamedIndividual'])
        else:
            individuals = self.graph.subjects(
                RDF['type'], uri)

        return list(individuals)

    def get_attributes(self):
        attributes = dict()
        # For each ObjectProperty found out corresponding range
        ranges = dict()

        parent_ranges = dict()

        restrictions = dict()

        # Check owl restrictions on classes
        for class_name in self.graph.subjects(RDF['type'], OWL['Class']):
            if not self.is_deprecated(class_name):
                for class_restr in self.graph.objects(
                        class_name, RDFS['subClassOf']):
                    if isinstance(class_restr, term.BNode):
                        for prp in self.graph.objects(
                                class_restr, OWL['onProperty']):
                            attributes.setdefault(class_name, set()).add(prp)
                            for child_class in self.graph.transitive_subjects(
                                    RDFS['subClassOf'], class_name):
                                if not self.is_deprecated(child_class):
                                    attributes.setdefault(
                                        child_class, set()).add(prp)

        # Attributes that can be found in all classes
        for prp, p, o in self.graph.triples((None, RDF['type'], None)):
            if not self.is_deprecated(prp):
                if o == OWL['DatatypeProperty'] or o == OWL['ObjectProperty']:
                    for class_name in self.graph.objects(prp, RDFS['domain']):
                        if not self.is_deprecated(class_name):
                            # Add attribute to current class
                            attributes.setdefault(class_name, set()).add(prp)

                        # Add attribute to children of current class
                        for child_class in self.graph.transitive_subjects(
                                RDFS['subClassOf'], class_name):
                            if not self.is_deprecated(child_class):
                                # Add attribute to current class
                                attributes.setdefault(
                                    child_class, set()).add(prp)

                    for range_name in self.graph.objects(prp, RDFS['range']):
                        # More complex type including restrictions
                        if isinstance(range_name, term.BNode):
                            for restriction_node in self.graph.objects(
                                    range_name, OWL['withRestrictions']):
                                for first_restriction in self.graph.objects(
                                        restriction_node, RDF['first']):
                                    xsd_restrictions = set(
                                        ['minInclusive', 'minExclusive',
                                         'maxInclusive', 'maxExclusive'])
                                    for rct in xsd_restrictions:
                                        for min_incl in self.graph.objects(
                                                first_restriction,
                                                XSD[rct]):
                                            if (prp in restrictions):
                                                if (rct in restrictions[prp]):
                                                    restrictions[prp] = max(
                                                        restrictions[prp][rct],
                                                        min_incl)
                                                else:
                                                    restrictions[prp] = {
                                                        rct: min_incl}
                                            else:
                                                restrictions[prp] = {
                                                    rct: min_incl}

                            for sub_range_name in self.graph.objects(
                                    range_name, OWL['onDatatype']):
                                range_name = sub_range_name

                        parent_ranges.setdefault(prp, set()).add(range_name)

                        # Add child_class to range (for ObjectProperty)
                        for child in self.graph.transitive_subjects(
                                RDFS['subClassOf'], range_name):
                            if not self.is_deprecated(child):
                                ranges.setdefault(prp, set()).add(child)

        return list((attributes, ranges, restrictions, parent_ranges))

    def get_graph(self):
        # Read owl (turtle) file
        owl_graph = Graph()
        # if self.file[0:4] == "http":
        #     owl_txt = urllib2.urlopen(self.file).read()
        # else:
        #     owl_txt = open(self.file, 'r').read()
        owl_graph.parse(self.file, format='turtle')
        # owl_graph.parse(data=owl_txt, format='turtle')

        if self.import_files:
            for import_file in self.import_files:
                # Read owl (turtle) file
                import_graph = Graph()
                # if self.file[0:4] == "http":
                #     import_txt = urllib2.urlopen(import_file).read()
                # else:
                #     import_txt = open(import_file, 'r').read()

                # This is a workaround to avoid issue with "#" in base prefix
                # as described in https://github.com/RDFLib/rdflib/issues/379,
                # When the fix is introduced in rdflib these 2 lines will be
                # replaced by:
                import_graph.parse(import_file, format='turtle')
                # import_txt = import_txt.replace(
                #     "http://www.w3.org/2002/07/owl#",
                #     "http://www.w3.org/2002/07/owl")
                # import_graph.parse(data=import_txt, format='turtle')

                owl_graph = owl_graph + import_graph

        return owl_graph

    def get_prov_class(self, owl_term, recursive=3):
        if not isinstance(owl_term, term.BNode) and \
                self.graph.qname(owl_term).startswith("prov:"):
            prov_class = owl_term
        else:
            parent_classes = list(self.graph.objects(
                owl_term, RDFS['subClassOf']))

            prov_class = None
            for parent_class in parent_classes:
                if not isinstance(parent_class, term.BNode):
                    if self.graph.qname(parent_class).startswith("prov:"):
                        prov_class = parent_class
                        break

            if not prov_class:
                if recursive > 0:
                    parent_classes_super = parent_classes
                    for parent_class_super in parent_classes_super:
                        if not isinstance(parent_class_super, term.BNode):
                            prov_class = self.get_prov_class(
                                parent_class_super, recursive=recursive-1)
                            if prov_class is not None:
                                break

        # Get mor generic PROV types
        if prov_class in set([PROV['SoftwareAgent'], PROV['Organization'],
                              PROV['Person']]):
            prov_class = PROV['Agent']
        if prov_class in set([PROV['Collection'], PROV['Location'],
                              PROV['Bundle'], PROV['Plan'],
                              PROV['EmptyCollection']]):
            prov_class = PROV['Entity']
        if prov_class in set([
                PROV['Usage'], PROV['AgentInfluence'],
                PROV['Derivation'], PROV['PrimarySource'],
                PROV['ActivityInfluence'], PROV['Delegation'],
                PROV['Start'], PROV['EntityInfluence'], PROV['Quotation'],
                PROV['End'], PROV['Attribution'], PROV['Association'],
                PROV['Revision'], PROV['Communication'],
                PROV['Generation']]):
            prov_class = PROV['Influence']
        if prov_class == PROV['Invalidation']:
            prov_class = PROV['InstantaneousEvent']

        return prov_class

    def get_definition(self, owl_term, add_links=True):
        definition = list(self.graph.objects(owl_term, OBO_DEFINITION))
        definition = definition + \
            list(self.graph.objects(owl_term, SKOS_DEFINITION))

        if definition:
            if len(definition) > 1:
                warnings.warn('Multiple definitions for '
                              + self.get_label(owl_term) + ': '
                              + ",".join(definition))
            definition = unicode(definition[0])
        else:
            definition = ""

        # Add link to term in document if the definition refer to a term
        if add_links:
            # definition = re.sub(
                # "("+"|".join(self.labels.keys())+")", "[\\1]", definition)
            terms = re.findall(r'\'.*?\'', definition)
            for mterm in sorted(set(terms), key=len, reverse=True):
                literal = Literal(mterm.replace("'", ""))
                if str(literal) in self.labels:
                    purl = self.labels[str(literal)]
                    if "#" in purl and \
                            not self.is_deprecated(term.URIRef(mterm)):
                        definition = definition.replace(
                            mterm,
                            "<a title=" +
                            purl.split("#")[1] + ">" +
                            mterm.replace("[", "").replace("]", "")+"</a>")

        # Remove final dot if present
        if definition:
            if definition[-1] == ".":
                definition = definition[:-1]

        return definition

    def get_individual_type(self, owl_term):
        indiv_type = list(self.graph.objects(owl_term, RDF['type']))
        if OWL['NamedIndividual'] in indiv_type:
            indiv_type.remove(OWL['NamedIndividual'])

        if indiv_type:
            indiv_type = ", ".join(map(self.graph.qname, indiv_type))
        else:
            indiv_type = ""
        return indiv_type

    def get_same_as(self, owl_term):
        same_as = list(self.graph.objects(owl_term, OWL['sameAs']))
        if same_as:
            same_as = ", ".join(same_as)
        else:
            same_as = ""
        return same_as

    def get_used_by(self, owl_term):
        used_by = list(self.graph.objects(owl_term, PROV['used']))
        return used_by

    def get_generated_by(self, owl_term):
        generated_by = list(self.graph.objects(
            owl_term, PROV['wasGeneratedBy']))
        return generated_by

    def get_example(self, owl_term, base_repository=None):
        example_list = list()

        examples = list(self.graph.objects(owl_term, OBO_EXAMPLE))

        for example in examples:
            if (base_repository is not None) and \
                    example.startswith(base_repository):
                local_path = example.replace(base_repository, "./")
                fid_ex = open(local_path)
                example = fid_ex.read()
                fid_ex.close()
            elif example.startswith("http"):
                with vcr.use_cassette(
                        os.path.join(NIDM_PATH, 'vcr_cassettes/synopsis.yaml'),
                        record_mode='new_episodes'):
                    # Read file from url
                    example = urllib2.urlopen(example).read()

            title = ""
            if example.startswith("#"):
                title = (example.split('\n', 1)[0]).replace("#", "")
                example = example.replace("#" + title, "")

            example_list.append([title, example])

        return sorted(example_list)

    def get_range(self, owl_term):
        ranges = list(self.graph.objects(owl_term, RDFS['range']))

        range_display = ""

        for range_value in ranges:
            if not isinstance(range_value, term.BNode):
                if isinstance(range_value, term.URIRef):
                    range_display += str(self.graph.qname(range_value))+" "
                else:
                    range_display += str(range_value)
        return range_display

    def get_domain(self, owl_term):
        domains = list(self.graph.objects(owl_term, RDFS['domain']))

        domain_display = ""

        for domain_value in sorted(domains):
            if isinstance(domain_value, term.URIRef):
                domain_display += str(self.graph.qname(domain_value))+" "
            else:
                domain_display += str(domain_value)
        return domain_display

    def get_curation_status(self, owl_term):
        curation_status = list(self.graph.objects(
            owl_term, HAS_CURATION_STATUS))
        if curation_status:
            if len(curation_status) > 1:
                warnings.warn('Multiple curation status for '
                              + self.get_label(owl_term) + ': '
                              + ",".join(curation_status))
            curation_status = curation_status[0]
        else:
            # By default consider that term is "uncurated"
            curation_status = OBO_UNCURATED
        return curation_status

    def get_editor(self, owl_term):
        editor = list(self.graph.objects(owl_term, OBO_TERM_EDITOR))
        if editor:
            if len(editor) > 1:
                warnings.warn('Multiple editors for '
                              + self.get_label(owl_term) + ': '
                              + ",".join(editor))
            editor = " (editor: "+editor[0]+")"
        else:
            editor = ""
        return editor

    def get_editor_note(self, owl_term):
        editor_note = list(self.graph.objects(owl_term, OBO_EDITOR_NOTE))
        if editor_note:
            if len(editor_note) > 1:
                warnings.warn('Multiple editor notes for '
                              + self.get_label(owl_term) + ': '
                              + ",".join(editor_note))
            editor_note = editor_note[0]
        else:
            editor_note = ""
        return editor_note

    def get_sub_class_names(self):
        sub_types = set()

        prov_types = set([PROV['Entity'], PROV['Activity'], PROV['Agent']])
        for prov_type in prov_types:
            for instance_id in self.graph.subjects(RDF.type, prov_type):
                for class_name in self.graph.objects(instance_id, RDF.type):
                    if not class_name == prov_type:
                        sub_types.add(class_name)

        return sub_types

    def check_class_names(self, ex_graph, ex_name, raise_now=False):
        my_exception = dict()
        error_msg = ""
        class_names = self.get_class_names()

        unrecognised_classes = list()
        # for class_name in ex_graph.subjects(RDF['type'], OWL['Class']):
        for class_name in ex_graph.objects(None, RDF.type):
            if not isinstance(class_name, term.BNode):
                if not class_name in class_names:
                    unrecognised_classes.append(class_name)

        for unrecognised_class in unrecognised_classes:
            if not unrecognised_class.startswith(str(PROV)):
                # key = ex_graph.qname(not_recognised_sub_type)
                key = "\n Unrecognised sub-type: " + \
                    ex_graph.qname(unrecognised_class)
                error_msg += key
                if key in my_exception:
                    my_exception[key].add(ex_name)
                else:
                    my_exception[key] = set([ex_name])

        if raise_now:
            if error_msg:
                raise Exception(error_msg)

        return my_exception

    def check_attributes(self, ex_graph, ex_name, raise_now=False):
        ignored_attributes = set([
            RDFS['label'],
            RDF['type'],
            PROV['value'], PROV['atTime'], PROV['used'],
            PROV['wasAssociatedWith'],
            PROV['qualifiedGeneration'], PROV['wasGeneratedBy'],
            PROV['atLocation'],
            PROV['wasAttributedTo'],
            PROV['activity'],
            PROV['wasDerivedFrom'],
            CRYPTO['sha512']
            ])

        my_exception = dict()
        my_range_exception = dict()
        my_restriction_exception = dict()

        owl_attributes = self.attributes
        owl_ranges = self.ranges
        owl_restrictions = self.type_restrictions

        # Find all attributes
        for s, p, o in ex_graph.triples((None, None, None)):
            # To be a DataTypeProperty then o must be a literal
            # if isinstance(o, rdflib.term.Literal):
            if p not in ignored_attributes:
                # *** Check domain
                # Get all defined types of current object
                found_attributes = False
                class_names = ""
                for class_name in sorted(ex_graph.objects(
                        s, RDF['type'])):
                    attributes = owl_attributes.get(class_name)

                    # If the current class was defined in the owl file check if
                    # current attribute was also defined.
                    if attributes:
                        if p in attributes:
                            found_attributes = True

                    class_names += ", "+ex_graph.qname(class_name) + \
                                   " (i.e. " + self.get_label(class_name) + ")"

                # if not found_attributes:
                    # if attributes:
                        # if not (p in attributes):
                if not found_attributes:
                    key = "\n Unrecognised attribute: " + \
                          ex_graph.qname(p) + \
                          " (i.e. " + self.get_label(p) + ")" + \
                          " in "+class_names[2:]
                    if not key in my_exception:
                        my_exception[key] = set([ex_name])
                    else:
                        my_exception[key].add(ex_name)

                # *** Check range for ObjectProperties and DataProperties
                if isinstance(o, term.URIRef):
                    # An ObjectProperty can point to an instance, then we look
                    # for its type:
                    found_range = set(ex_graph.objects(o, RDF['type']))

                    # An ObjectProperty can point to a term
                    if not found_range:
                        found_range = set([o])

                        # If the term is an individual, look for its type
                        if OWL['NamedIndividual'] in \
                                set(self.graph.objects(o, RDF['type'])):
                            found_range = set(
                                self.graph.objects(o, RDF['type']))

                elif isinstance(o, term.Literal):
                    found_range = set([o.datatype])

                correct_range = False
                if p in owl_ranges:
                    # If none of the class found for current ObjectProperty
                    # value is part of the range throw an error

                    # If the type of current value is within the authorised
                    # ranges
                    if found_range.intersection(owl_ranges[p]):
                        correct_range = True
                    else:
                        if p in owl_ranges:
                            # A bit more complicated to deal with
                            # "positiveInteger"
                            for owl_range in owl_ranges[p]:
                                # FIXME: we should be able to do better than
                                # that to check that XSD['positiveInteger'] is
                                # in owl_ranges[p]
                                if (XSD['positiveInteger'] == owl_range) \
                                        and (next(iter(found_range)) ==
                                             XSD['int']) \
                                        and (o.value >= 0):
                                    correct_range = True
                    if not correct_range:
                        found_range_line = ""
                        # FIXME: This should be better handled to be able to do
                        # "if found_range"
                        if not None in found_range:
                            found_range_line = ', '.join(
                                map(self.get_name_label, sorted(found_range)))
                        owl_range_line = ""
                        if p in owl_ranges:
                            owl_range_line = ', '.join(
                                map(self.get_name_label, sorted(owl_ranges[p]))
                                )

                        key = "\n Unrecognised range: " + \
                            found_range_line + \
                            ' for '+self.get_name_label(p) + \
                            ' should be ' + \
                            owl_range_line
                else:
                    # No range found for current attribute
                    correct_range = False
                    key = "\n No range defined for: " + ex_graph.qname(p)

                if not correct_range:
                    if not key in my_range_exception:
                        my_range_exception[key] = set([ex_name])
                    else:
                        my_range_exception[key].add(ex_name)

                if p in owl_restrictions:
                    restrictions_ok = True
                    if 'minInclusive' in owl_restrictions[p]:
                        if o.value < owl_restrictions[p]['minInclusive'].value:
                            restrictions_ok = False
                    if 'minExclusive' in owl_restrictions[p]:
                        if o.value <= \
                                owl_restrictions[p]['minExclusive'].value:
                            restrictions_ok = False
                    if 'maxInclusive' in owl_restrictions[p]:
                        if o.value > owl_restrictions[p]['maxInclusive'].value:
                            restrictions_ok = False
                    if 'maxExclusive' in owl_restrictions[p]:
                        if o.value >= \
                                owl_restrictions[p]['maxExclusive'].value:
                            restrictions_ok = False
                    if not restrictions_ok:
                        key = "\n Contraints: value "+str(o.value) + \
                            ' for '+ex_graph.qname(p) + \
                            ' does not observe contraints ' + \
                            ', '.join(sorted(owl_restrictions[p]))
                        if not key in my_restriction_exception:
                            my_restriction_exception[key] = set([ex_name])
                        else:
                            my_restriction_exception[key].add(ex_name)

        if raise_now:
            error_msg = ""
            for exc in my_exception.keys():
                error_msg += exc
            for exc in my_range_exception.keys():
                error_msg += exc
            for exc in my_restriction_exception.keys():
                error_msg += exc

            if error_msg:
                raise Exception(error_msg)

        return list((my_exception, my_range_exception,
                     my_restriction_exception))

    def get_label(self, uri):
        if not isinstance(uri, term.BNode):
            try:
                name = self.graph.qname(uri)
            except:
                # For ontology names, qname fails not sure if this is a bug
                name = uri
        else:
            name = uri

        # If a label is available, use the namespace:label, otherwise qname
        label = list(self.graph.objects(uri, RDFS['label']))
        if label:
            if len(label) > 1 and not self.is_external_namespace(uri):
                warnings.warn('Multiple labels for '+name+': '+",".join(label))
            label = sorted(label)[0]
            name = name.split(":")[0]+":'"+label+"'"

        return name

    def is_external_namespace(self, term_uri):
        return not (term_uri.startswith(NIDM)
                    or term_uri.startswith(FSL)
                    or term_uri.startswith(SPM)
                    or term_uri.startswith(AFNI))

    def is_prov(self, term_uri):
        term_label = self.get_label(term_uri)

        return term_label.startswith("prov")

    def get_name(self, uri):
        name = self.graph.qname(uri).split(":")[1]
        return name

    def get_name_label(self, uri):
        if self.get_name(uri) == self.get_label(uri):
            return self.get_name(uri)
        else:
            return self.get_name(uri) + " (i.e. " + self.get_label(uri) + ")"

    def get_preferred_prefix(self, uri):
        label = str(self.graph.label(uri))
        idt = str(self.graph.qname(uri).split(":")[1])

        if label == idt:
            # For text identifiers there is no preferred prefix (we can just
            # keep the id directly)
            prefix_name = None
        else:
            prefix_name = self.get_label(uri).replace(" ", "")\
                                             .replace(":", "_")\
                                             .replace("'", "")\
                                             .replace("-", "")
        return prefix_name

    def sorted_by_labels(self, term_list):
        class_labels = map(self.get_label, term_list)
        sorted_term_list = [x for (y, x) in
                            sorted(zip(class_labels, term_list))]

        return sorted_term_list

    def prefixes_as_csv(self, csvfile):
        with open(csvfile, 'wb') as fid:
            writer = csv.writer(fid)
            writer.writerow(["qname", "Preferred prefix", "URI"])

            # For anything that has a label
            for s, o in sorted(self.graph.subject_objects(RDFS['label'])):
                try:
                    self.graph.qname(s)
                except:
                    # Some URIs don't have qname
                    # (e.g. http://www.w3.org/ns/prov-o#)
                    continue
                prefix = self.get_preferred_prefix(s)

                if prefix is not None and not self.is_deprecated(s):
                    writer.writerow([self.graph.qname(s), prefix, s])



from setuptools import setup, find_packages

readme = open('README.rst').read()

reqs = [line.strip() for line in open('requirements.txt').readlines()]
requirements = list(filter(None, reqs))

setup(
    name="nidmresults",
    version="0.3.4",
    author="Camille Maumet",
    author_email="c.m.j.maumet@warwick.ac.uk",
    description=(
        "Export of neuroimaging statistical results using NIDM"
        " as specified at http://nidm.nidash.org/specs/nidm-results.html."),
    license = "MIT",
    scripts=['bin/nidmreader', 'bin/nidm_mkda_convert', 'bin/nidmresults'],
    url='https://github.com/incf-nidash/nidmresults',
    keywords = "Prov, NIDM, Provenance",
    packages=find_packages(),
    package_dir={
        'nidmresults': 'nidmresults'
    },
    long_description=readme,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
    package_data={'':
                  ['prefixes.csv',
                   'owl/nidm-results_020.owl',
                   'owl/nidm-results_100.owl',
                   'owl/nidm-results_110.owl',
                   'owl/nidm-results_120.owl',
                   'owl/nidm-results_130.owl']},
    include_package_data=True,
    install_requires=requirements,
)
